{
module Syntax.Lexer where

import Data.List

import Syntax.ParseMonad
import Syntax.Position

import Utils.Debug

}

$digit	    = 0-9
$alphanum   = [A-Za-z0-9_']
$idstart    = [A-Za-z]
$symstart   = [\!\#\$\%\&\*\+\.\/\<\=\>\?\@\\\^\|\-\~\:]
$symbol	    = [\!\#\$\%\&\*\+\.\/\<\=\>\?\@\\\^\|\-\~\:']

$white_nonl = $white # \n

tokens :-

<tex>	 ^ \\ "begin{code}" \n	{ begin_ code }
<tex>	 ^ .* \n		{ withRange $ \_ -> TkTeX }
<tex>	 ^ .+			{ withRange $ \_ -> TkTeX }
<code>   ^ \\ "end{code}" \n	{ end_ }

<0,code,bol,layout,layout_left>
    $white_nonl+	;

<0,code,bol,layout,layout_left>
    "{-"		{ nestedComment }

-- Dashes followed by an operator symbol should be parsed as an operator.
<0,code,bol,layout,layout_left>
    "--"\-* [^$symstart] .*   ;
<0,code,bol,layout,layout_left>
    "--"\-* / { atEOL }	    ;
<0,code> \n		{ begin bol }
<bol>
    {
	\n		;
	()		{ beginningOfLine }
    }
<layout>
    {	\{ / { notFollowedBy '-' }  { popAnd openBrace }
	\n			    ;
	()			    { newLayoutContext }
    }
<layout_left> ()		{ layoutLeft }

<0,code> Set			{ withRange_ TkSet }
<0,code> Prop			{ withRange_ TkProp }
<0,code> Type			{ withRange_ TkType }
<0,code> data			{ withRange_ TkData }
<0,code> sort			{ withRange_ TkSort }
<0,code> where			{ withLayout $ withRange_ TkWhere }
<0,code> let			{ withLayout $ withRange_ TkLet }
<0,code> in			{ withRange_ TkIn }
<0,code> infix			{ withRange_ TkInfix }
<0,code> infixl			{ withRange_ TkInfixL }
<0,code> infixr			{ withRange_ TkInfixR }
<0,code> $idstart $alphanum* \- plugin
				{ withRange $ \r s -> TkPlugin (r, takeWhile ('-'/=) s) }
<0,code> $idstart $alphanum*	{ withRange (curry TkId) }
<0,code> \_ $alphanum+		{ withRange (curry TkId) }
<0,code> $digit+		{ withRange $ \r s -> TkInt (r,read s) }
<0,code> \_			{ withRange_ TkUnderscore }    
<0,code> \#$digit+		{ withRange $ \r s -> TkUniverse (r, read (tail s)) }
<0,code> \(			{ withRange_ TkLParen }
<0,code> \)			{ withRange_ TkRParen }
<0,code> \\			{ withRange_ TkLambda }
<0,code> \-\>			{ withRange_ TkArrow }
<0,code> \,			{ withRange_ TkComma }
<0,code> \;			{ withRange_ TkSemi }
<0,code> \:\:			{ withRange_ TkDblColon }
<0,code> \=\>			{ withRange_ TkDblArrow }
<0,code> \=			{ withRange_ TkEqual }
<0,code> \.			{ withRange_ TkDot }
<0,code> \`			{ withRange_ TkBackQuote }
<0,code> \{			{ openBrace }
<0,code> \}			{ closeBrace }
<0,code> \[			{ withRange_ TkOpenSquare }
<0,code> \]			{ withRange_ TkCloseSquare }
<0,code> $symstart $symbol*	{ withRange (curry TkOp) }

{
data Token  = TkLambda Range	| TkArrow Range  | TkId (Range, String)
	    | TkLParen Range	| TkRParen Range | TkSemi Range | TkComma Range
	    | TkUniverse (Range,Int)
	    | TkSet Range	| TkType Range	 | TkProp Range
	    | TkInfix Range	| TkInfixL Range | TkInfixR Range
            | TkBackQuote Range	| TkWhere Range	 | TkData Range	| TkSort Range
	    | TkInt (Range, Int)| TkDblArrow Range
	    | TkOpenBrace Range	| TkCloseBrace Range
	    | TkOpenSquare Range| TkCloseSquare Range
	    | TkVOpenBrace Range| TkVCloseBrace Range	| TkVSemi Range
	    | TkEqual Range	| TkDblColon Range    | TkDot Range
	    | TkLet Range	| TkIn Range	    | TkPlugin (Range,String)
            | TkUnderscore Range| TkOp (Range, String)
	    | TkTeX String
	    | TkEOF
    deriving Show

scanToken :: Parser Token
scanToken =
    do	inp <- getLexInput
	ls:_ <- getLexState
	case alexScan inp ls of
	    AlexEOF	    -> scanEOF inp TkEOF
	    AlexError inp'  -> scanError inp'
	    AlexSkip inp' _ -> scanSkip inp' scanToken
	    AlexToken inp' len action
			    -> action inp inp' len
		

lexer k = scanToken >>= k

-- States

begin :: LexState -> LexAction Token
begin code _ _ _ =
    do	pushLexState code
	scanToken

begin_ :: LexState -> LexAction Token
begin_ code _ inp' _ =
    do	pushLexState code
	scanSkip inp' scanToken

end_ :: LexAction Token
end_ _ inp' _ =
    do	popLexState
	scanSkip inp' scanToken

popAnd :: LexAction a -> LexAction a
popAnd a inp inp' n =
    do	popLexState
	a inp inp' n

-- Layout

openBrace :: LexAction Token
openBrace = scanTok $ \_ ->
    do	pushContext NoLayout
	r <- getParseRange
	return (TkOpenBrace r)

closeBrace :: LexAction Token
closeBrace = scanTok $ \_ ->
    do	popContext
	r <- getParseRange
	return (TkCloseBrace r)

beginningOfLine :: LexAction Token
beginningOfLine (p,_,_) _ _ =
    do	offs <- getOffside p
	case offs of
	    LT	-> do	popContext
			return (TkVCloseBrace (Range p p))
	    EQ	-> do	popLexState
			return (TkVSemi (Range p p))
	    GT	-> do	popLexState
			scanToken

layoutLeft :: LexAction Token
layoutLeft (p,_,_) _ _ =
    do	popLexState
	pushLexState bol
	return (TkVCloseBrace (Range p p))

newLayoutContext :: LexAction Token
newLayoutContext (p,_,_) _ _ =
    do	popLexState
	let offset = posCol p
	ctx <- getContext
	case ctx of
	    Layout prev_off : _ | prev_off >= offset ->
		do  -- token is indented to the left of the previous context.
		    -- we must generate a {} sequence now.
		    pushLexState layout_left
		    return (TkVOpenBrace (Range p p))
	    _ -> do setContext (Layout offset : ctx)
		    return (TkVOpenBrace (Range p p))

notFollowedBy :: Char -> AlexInput -> AlexInput -> Int -> AlexInput -> Bool
notFollowedBy c _ _ _ (_,i,_) =
    null i || head i /= c

atEOL _ _ _ (_,i,c) = null i || c == '\n'

isBeginCode _ _ _ (_,i,_) = null i || ("\\begin{code}" `isPrefixOf` i)

withLayout :: LexAction r -> LexAction r
withLayout a i1 i2 n =
    do	pushLexState layout
	a i1 i2 n

-- Nested comments


-- Nested comments require traversing by hand, they can't be parsed
-- using regular expressions.
nestedComment :: LexAction Token
nestedComment (p,_,_) i1 n =
	go 1 i1
    where
	go 0 input = do	setLexInput input
			scanToken
	go n input =
	    case alexGetChar input of
		Nothing		-> err
		Just (c,input)	->
		    case c of
			'-' -> case alexGetChar input of
			    Nothing		-> err
			    Just ('}', input)	-> go (n-1) input
			    Just (c, _)         -> go n input
			'{' -> case alexGetChar input of
			    Nothing		-> err
			    Just ('-',input')	-> go (n+1) input'
			    Just (c,input)	-> go n input
			c -> go n input

        err = parseErrorAt p "Unterminated `{-'"

}
